// -------------------------------------------------------------
// 3d code by Georgy Moshkin (Delphi)
// -------------------------------------------------------------
unit matrice;

interface

uses vector;

type TMatrice = array[0..3,0..3] of single;

function Mat(a00,a01,a02,a03,
             a10,a11,a12,a13,
             a20,a21,a22,a23,
             a30,a31,a32,a33: single):TMatrice;

function RotMatX(A:single):TMatrice;
function RotMatY(A:single):TMatrice;
function RotMatZ(A:single):TMatrice;
function TraMat(x,y,z:single):TMatrice;
function ConcatMatrix(const Ma: TMatrice; const Mb: TMatrice):TMatrice;
Procedure ApplyMatrix(Var Vert: TVector; const M: TMatrice);
Procedure InvApplyMatrix(Var Vert: TVector; const M: TMatrice);

implementation

function Mat(a00,a01,a02,a03,
             a10,a11,a12,a13,
             a20,a21,a22,a23,
             a30,a31,a32,a33: single):TMatrice;
begin

result[0,0]:=a00;
result[0,1]:=a01;
result[0,2]:=a02;
result[0,3]:=a03;
result[1,0]:=a10;
result[1,1]:=a11;
result[1,2]:=a12;
result[1,3]:=a13;
result[2,0]:=a20;
result[2,1]:=a21;
result[2,2]:=a22;
result[2,3]:=a23;
result[3,0]:=a30;
result[3,1]:=a31;
result[3,2]:=a32;
result[3,3]:=a33;
end;

function RotMatX(A:single):TMatrice;
var cosA,sinA:single;
begin
cosA:=cos(A);
sinA:=sin(A);
result:= mat(1, 0,    0,     0,
             0, cosA, -sinA, 0,
             0, sinA,  cosA, 0,
             0, 0,     0,    1);
end;

function RotMatY(A:single):TMatrice;
var cosA,sinA:single;
begin
cosA:=cos(A);
sinA:=sin(A);
result:= mat(cosA, 0, sinA, 0,
	     0,    1,  0,    0,
	     -sinA,0,  cosA, 0,
	     0,    0,  0,    1);
end;

function RotMatZ(A:single):TMatrice;
var cosA,sinA:single;
begin
cosA:=cos(A);
sinA:=sin(A);
result:= mat(cosA, -sinA, 0, 0,
	     sinA,  cosA, 0, 0,
	     0,     0,    1, 0,
	     0,     0,    0, 1);

end;

function TraMat(x,y,z:single):TMatrice;
begin
result:=mat(1,0,0,x,
            0,1,0,y,
            0,0,1,z,
            0,0,0,1);
end;

function ConcatMatrix(const Ma: TMatrice; const Mb: TMatrice):TMatrice;
Var row,clm: integer;
begin
  for row := 0 to 3 do
   for clm := 0 to 3 do
    Result[row,clm] := ma[row,0] * mb[0,clm] +
                       ma[row,1] * mb[1,clm] +
                       ma[row,2] * mb[2,clm] +
                       ma[row,3] * mb[3,clm];
end;




Procedure ApplyMatrix(Var Vert: TVector; const M: TMatrice);
Var N: TVector;
begin
  N.x := vert.x * M[0,0] + Vert.y * M[0,1] + Vert.z * M[0,2] + M[0,3];
  N.y := vert.x * M[1,0] + Vert.y * M[1,1] + Vert.z * M[1,2] + M[1,3];
  N.z := vert.x * M[2,0] + Vert.y * M[2,1] + Vert.z * M[2,2] + M[2,3];
  Vert:=N;
end;


Procedure InvApplyMatrix(Var Vert: TVector; const M: TMatrice);
Var N: TVector;
begin
  vert.x := vert.x -  M[0,3];
  vert.y := vert.y -  M[1,3];
  vert.z := vert.z -  M[2,3];
  N.x := vert.x * M[0,0] + Vert.y * M[1,0] + Vert.z * M[2,0];
  N.y := vert.x * M[0,1] + Vert.y * M[1,1] + Vert.z * M[2,1];
  N.z:=  vert.x * M[0,2] + Vert.y * M[1,2] + Vert.z * M[2,2];
  vert:=N;
end;



end.
