// -------------------------------------------------------------
// 3d code by Georgy Moshkin (Delphi)
// -------------------------------------------------------------
unit skeleton;

interface

uses vector,matrice,dialogs,math;

type TBonePos=record
               trans  : TVector;
               rot    : TVector;
               absmat : TMatrice;
               relmat : TMatrice;
               point  : TVector;
              end;

type TSkeletonNode=record
                    name       : string;
                    parent     : integer;
                   end;

type TSkeleton=array of TSkeletonNode;

type TAnimFrame=array of TBonePos;

type TAnimFrames=array of TAnimFrame;

type TAction=record
              name       : string;
              fps        : single;
              animFrames : TAnimFrames;
             end;

type TActions=array of TAction;

type TSkeletonState=record
                     prevFrame:integer;
                     nextFrame:integer;
                     prevAction:integer;
                     nextAction:integer;
                     skelTime:single;
                     currentPose: TAnimFrame;
                    end;

type TBoneWeight=record
                  boneIdx: integer;
                  name   : string;
                  weight : single;
                 end;

type TVertParents=array of TBoneWeight;

type TVertsParents=array of TVertParents;


type TSkeletonAnimation=record
                         skeleton : TSkeleton;
                         actions  : TActions;
//                         skeletonState : TSkeletonState;
//                         vertsParents : TVertsParents;
                         reference  : TAnimFrame;
                        end;

procedure Anim(actions:TActions;var state:TSkeletonState;actionNum:integer;BlenderElapsed:single);

procedure Transform(var verts:TVectors; locverts:TVectors;vparents:TVertsParents;pose,restpose:TAnimFrame;skel:TSkeleton);
function IndexFromName(skel:TSkeleton;name:string):integer;
procedure BuildSkeleton(verts:TVectors;var locverts:TVectors;vparents:TVertsParents;var pose:TAnimFrame;skel:TSkeleton);
procedure BuildFrame(var Pose:TAnimFrame; skel:TSkeleton);

implementation

procedure Anim(actions:TActions;var state:TSkeletonState;actionNum:integer;BlenderElapsed:single);
var delta:single;
var i,k:integer;
    db:TVector;    
begin

with state do
 begin

  if prevAction<>ActionNum then nextFrame:=0{1};

  nextAction:=ActionNum;
  if prevAction=-1 then prevAction:=ActionNum;
  if nextAction=-1 then nextAction:=ActionNum;

  delta:= BlenderElapsed * Actions[nextAction].fps / 1000;

  skelTime:=skelTime + delta;

  if skelTime>1 then
   begin
    skelTime:=0;
    prevAction:=nextAction;
    prevFrame:=nextFrame;
    nextFrame:=nextFrame+1; //aDelta;

    if nextFrame>length(Actions[nextAction].animFrames)-1 then
     begin
      prevAction:=nextAction; // HACK
      nextFrame:=0; // in old versions 0 - reference.
     end


   end;


 // some hack
 if length(currentPose)=0 then setlength(currentPose,length(Actions[prevAction].animFrames[prevFrame]));

 for i:=0 to length(currentPose)-1 do
  begin

    currentpose[i].trans:=vectorint(Actions[prevAction].animFrames[prevFrame][i].trans,
                                   Actions[nextAction].animFrames[nextFrame][i].trans,
                                   skelTime);

    db.x:=zangle(Actions[prevAction].animFrames[prevFrame][i].rot.x,
                 Actions[nextAction].animFrames[nextFrame][i].rot.x);

    db.y:=zangle(Actions[prevAction].animFrames[prevFrame][i].rot.y,
                 Actions[nextAction].animFrames[nextFrame][i].rot.y);

    db.z:=zangle(Actions[prevAction].animFrames[prevFrame][i].rot.z,
                 Actions[nextAction].animFrames[nextFrame][i].rot.z);

  with Actions[prevAction].animFrames[prevFrame][i] do
   currentpose[i].rot:=v1plusv2(Actions[prevAction].animFrames[prevFrame][i].rot,
                                db,
                                skelTime);
(*
   currentpose[i].rot:=Actions[prevAction].animFrames[prevFrame][i].rot;
*)
  end;


 end;
end;

procedure Transform(var verts:TVectors; locverts:TVectors;vparents:TVertsParents;pose,restpose:TAnimFrame;skel:TSkeleton);
var i,k:integer;
    zzz,fff:TVector;
    mmm:TMatrice;
    w:single;

begin

for i:=0 to length(verts)-1 do
 begin
  fff.x:=0;
  fff.y:=0;
  fff.z:=0;
 for k:=0 to length(vparents[i])-1 do
  begin
   zzz:=locverts[i];

   mmm:=restpose[vparents[i][k].boneIdx].absmat;

   InvApplyMatrix(zzz,mmm);

   mmm:=pose[vparents[i][k].boneIdx].absmat;

   ApplyMatrix(zzz,mmm);   

   w:=vparents[i][k].weight;

   fff.x:=fff.x+zzz.x*w;
   fff.y:=fff.y+zzz.y*w;
   fff.z:=fff.z+zzz.z*w;
  end;
  verts[i]:=fff;
 end;



end;



function IndexFromName(skel:TSkeleton;name:string):integer;
var i:integer;
    found:boolean;
begin
found:=false;
 for i:=0 to length(skel)-1 do
  if skel[i].name=name then
   begin
    result:=i;
    found:=true;
   end;

if not found then showmessage('!');
end;

procedure BuildSkeleton(verts:TVectors;var locverts:TVectors;vparents:TVertsParents;var pose:TAnimFrame;skel:TSkeleton);
var i,k:integer;
    zzz,fff:TVector;
    mmm:TMatrice;
    w:single;
begin

BuildFrame(pose,skel);

setlength(locverts,length(verts));

for i:=0 to length(verts)-1 do
  locverts[i]:=verts[i];

for i:=0 to length(verts)-1 do
 for k:=0 to length(vparents[i])-1 do
   vparents[i][k].boneIdx:=IndexFromName(skel,vparents[i][k].name);

end;


procedure BuildFrame(var Pose:TAnimFrame; skel:TSkeleton);
var i:integer;
begin

for i:=0 to length(pose)-1 do
 begin
  pose[i].point.x:=0;
  pose[i].point.y:=0;
  pose[i].point.z:=0;
  pose[i].relmat:=rotMatX(pose[i].rot.x);
  pose[i].relmat:=ConcatMatrix(rotMatY(pose[i].rot.y),pose[i].relmat);
  pose[i].relmat:=ConcatMatrix(rotMatZ(pose[i].rot.z),pose[i].relmat);

  pose[i].relmat[0,3]:=pose[i].trans.x;
  pose[i].relmat[1,3]:=pose[i].trans.y;
  pose[i].relmat[2,3]:=pose[i].trans.z;

       if skel[i].parent<>-1 then
         pose[i].absmat:=ConcatMatrix(pose[skel[i].parent].absmat,
                                          pose[i].relmat)
         else
        pose[i].absmat:=pose[i].relmat;


 end;

  for i:=0 to length(pose)-1 do
   ApplyMatrix(pose[i].point,
               pose[i].absmat);




end;


end.
